/*
Copyright (C) 2011 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

package coordonnees;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Hashtable;
import javax.swing.BorderFactory;
import javax.swing.JApplet;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class AppletCoordonnees extends JApplet implements ChangeListener {
    
    static final double minAlpha = Math.toRadians(0);
    static final double maxAlpha = Math.toRadians(360);
    static final double minDelta = Math.toRadians(-90);
    static final double maxDelta = Math.toRadians(90);
    static final double minPhi = Math.toRadians(-90);
    static final double maxPhi = Math.toRadians(90);
    static final double minL = Math.toRadians(-180);
    static final double maxL = Math.toRadians(180);
    static final int minHeure = 0;
    static final int maxHeure = 23;
    static final int minJour = 1;
    int maxJour = 31; // dpend du mois
    static final int minMois = 1;
    static final int maxMois = 12;
    static final int minAnnee = 2000;
    static final int maxAnnee = 2020;
    static final double alphaInitial = Math.toRadians(56.25); // 3h 45m
    static final double deltaInitial = Math.toRadians(24); // +24
    static final double phiInitial = Math.toRadians(49); // latitude 49 Nord
    static final double LInitial = Math.toRadians(2); // longitude 2 Est
    static final int heureInitiale = 0;
    static final int jourInitial = 1;
    static final int moisInitial = 1;
    static final int anneeInitiale = 2000;
    static final Color couleurEquateur = new Color(150, 255, 200, 200);
    static final Color couleurTexteEquateur = new Color(100, 200, 150);
    static final Color couleurHorizon = new Color(250, 210, 170, 200);
    static final Color couleurTexteHorizon = new Color(200, 150, 100);
    static final Color couleurEtoile = new Color(255, 255, 0);
    static final Color couleurAlpha = new Color(255, 50, 50, 230);
    static final Color couleurDelta = new Color(50, 50, 255);
    static final Color couleurPhi = new Color(50, 130, 50);
    static final Color couleurArcs = Color.lightGray;
    static final Color couleurAngleA = new Color(150, 100, 50);
    static final Color couleurAngleH = new Color(50, 150, 100);
    static final Color couleurAngleHauteur = new Color(150, 50, 150);
    static final int precisionCourbes = 100;
    int imgdimx;
    int imgdimy;
    protected ImageCanvas ic;
    protected Image croquis = null;
    protected Graphics2D g2d = null;
    protected JSlider sliderAlpha, sliderDelta, sliderPhi, sliderL, sliderHeure, sliderJour, sliderMois, sliderAnnee;
    protected double alpha, delta, phi, L;
    protected int heure, jour, mois, annee;
    double H = Math.PI / 3; // angle horaire
    protected int cx, cy; // position du centre de l'image en pixels
    protected int rayon; // rayon du grand cercle en pixels
    protected JLabel labelAlpha, labelDelta, labelPhi, labelL, labelHeure, labelJour, labelMois, labelAnnee;
    protected JLabel labelInfos;
    protected Font gras;
    protected Matrice rotObservateur; // matrice de rotation par rapport  l'observateur
    protected Matrice rotEquateur; // matrice de rotation de l'quateur cleste par rapport  l'horizon
    protected Matrice rotObsEq; // rotObservateur * rotEquateur
    protected Vecteur O, Nord, Sud, Est, Ouest, Z; // points en coordonnes sphriques
    protected String titreAlpha, titreDelta, titrePhi, titreGamma;
    
    
    public void init() {
        imgdimx = Math.min(getWidth() - 200, getHeight() - 50);
        if (imgdimx <= 0)
            imgdimx = 500;
        imgdimy = imgdimx;
        getContentPane().setLayout(new BorderLayout());
        getContentPane().setBackground(Color.white);
        
        final JPanel croquisPane = new JPanel(new BorderLayout());
        croquisPane.setBackground(Color.white);
        creerCroquis();
        ic = new ImageCanvas(imgdimx, imgdimy);
        croquisPane.add(ic, BorderLayout.CENTER);
        labelInfos = new JLabel();
        croquisPane.add(labelInfos, BorderLayout.SOUTH);
        getContentPane().add(croquisPane, BorderLayout.WEST);
        
        final JPanel controlPane = new JPanel(new BorderLayout());
        controlPane.setBorder(BorderFactory.createEmptyBorder(0, 10, 0, 0));
        final JPanel labelPane = new JPanel(new GridLayout(8, 1));
        labelPane.setPreferredSize(new Dimension(120, imgdimy));
        labelPane.setBackground(Color.white);
        controlPane.add(labelPane, BorderLayout.WEST);
        final JPanel sliderPane = new JPanel(new GridLayout(8, 1));
        sliderPane.setBackground(Color.white);
        controlPane.add(sliderPane, BorderLayout.CENTER);
        controlPane.setBackground(Color.white);
        getContentPane().add(controlPane, BorderLayout.CENTER);
        
        alpha = alphaInitial;
        if (gras.canDisplay('\u03B1'))
            titreAlpha = "\u03B1";
        else
            titreAlpha = "alpha";
        labelAlpha = new JLabel(titreAlpha + " = " + affichageAngleHMS(alpha));
        labelAlpha.setForeground(couleurAlpha);
        labelAlpha.setFont(gras);
        labelPane.add(labelAlpha);
        sliderAlpha = new JSlider(0, 96, (int)Math.round(96 * (alpha - minAlpha) / (maxAlpha - minAlpha)));
        sliderAlpha.addChangeListener(this);
        sliderAlpha.setMajorTickSpacing(24);
        sliderAlpha.setPaintTicks(true);
        Hashtable graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel(""+Math.round(minAlpha*12/Math.PI)));
        graduations.put(new Integer(48), new JLabel(""+Math.round((minAlpha + (maxAlpha - minAlpha)/2)*12/Math.PI)));
        graduations.put(new Integer(96), new JLabel(""+Math.round(maxAlpha*12/Math.PI)));
        sliderAlpha.setLabelTable(graduations);
        sliderAlpha.setPaintLabels(true);
        sliderAlpha.setBackground(Color.white);
        sliderPane.add(sliderAlpha);
        
        delta = deltaInitial;
        if (gras.canDisplay('\u03B4'))
            titreDelta = "\u03B4";
        else
            titreDelta = "delta";
        labelDelta = new JLabel(titreDelta + " = " + affichageAngleDMS(delta, false, false));
        labelDelta.setForeground(couleurDelta);
        labelDelta.setFont(gras);
        labelPane.add(labelDelta);
        sliderDelta = new JSlider(0, 180, (int)Math.round(180 * (delta - minDelta) / (maxDelta - minDelta)));
        sliderDelta.addChangeListener(this);
        sliderDelta.setMajorTickSpacing(45);
        sliderDelta.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel(""+Math.round(Math.toDegrees(minDelta))));
        graduations.put(new Integer(90), new JLabel(""+Math.round(Math.toDegrees(minDelta + (maxDelta - minDelta)/2))));
        graduations.put(new Integer(180), new JLabel(""+Math.round(Math.toDegrees(maxDelta))));
        sliderDelta.setLabelTable(graduations);
        sliderDelta.setPaintLabels(true);
        sliderDelta.setBackground(Color.white);
        sliderPane.add(sliderDelta);
        
        phi = phiInitial;
        if (gras.canDisplay('\u03C6'))
            titrePhi = "\u03C6";
        else
            titrePhi = "phi";
        labelPhi = new JLabel(titrePhi + " = " + affichageAngleDMS(phi, false, false));
        labelPhi.setForeground(couleurPhi);
        labelPhi.setFont(gras);
        labelPane.add(labelPhi);
        sliderPhi = new JSlider(0, 180, (int)Math.round(180 * (phi - minPhi) / (maxPhi - minPhi)));
        sliderPhi.addChangeListener(this);
        sliderPhi.setMajorTickSpacing(45);
        sliderPhi.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel(""+Math.round(Math.toDegrees(minPhi))));
        graduations.put(new Integer(90), new JLabel(""+Math.round(Math.toDegrees(minPhi + (maxPhi - minPhi)/2))));
        graduations.put(new Integer(180), new JLabel(""+Math.round(Math.toDegrees(maxPhi))));
        sliderPhi.setLabelTable(graduations);
        sliderPhi.setPaintLabels(true);
        sliderPhi.setBackground(Color.white);
        sliderPane.add(sliderPhi);
        
        L = LInitial;
        labelL = new JLabel("L = " + affichageAngleDMS(Math.abs(L), true, false));
        labelL.setFont(gras);
        labelPane.add(labelL);
        sliderL = new JSlider(0, 180, (int)Math.round(180 * (L - minL) / (maxL - minL)));
        sliderL.addChangeListener(this);
        sliderL.setMajorTickSpacing(45);
        sliderL.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel(""+Math.round(Math.toDegrees(minL))));
        graduations.put(new Integer(90), new JLabel(""+Math.round(Math.toDegrees(minL + (maxL - minL)/2))));
        graduations.put(new Integer(180), new JLabel(""+Math.round(Math.toDegrees(maxL))));
        sliderL.setLabelTable(graduations);
        sliderL.setPaintLabels(true);
        sliderL.setBackground(Color.white);
        sliderPane.add(sliderL);
        
        heure = heureInitiale;
        labelHeure = new JLabel("Heure = " + heure + "h TU");
        labelHeure.setFont(gras);
        labelPane.add(labelHeure);
        sliderHeure = new JSlider(minHeure, maxHeure, heure);
        sliderHeure.setSnapToTicks(true);
        sliderHeure.addChangeListener(this);
        sliderHeure.setMajorTickSpacing(6);
        sliderHeure.setMinorTickSpacing(1);
        sliderHeure.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel("0"));
        graduations.put(new Integer(12), new JLabel("12"));
        sliderHeure.setLabelTable(graduations);
        sliderHeure.setPaintLabels(true);
        sliderHeure.setBackground(Color.white);
        sliderPane.add(sliderHeure);
        
        jour = jourInitial;
        labelJour = new JLabel("Jour = " + jour);
        labelJour.setFont(gras);
        labelPane.add(labelJour);
        sliderJour = new JSlider(minJour, maxJour, jour);
        sliderJour.setSnapToTicks(true);
        sliderJour.addChangeListener(this);
        sliderJour.setMinorTickSpacing(1);
        sliderJour.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(0), new JLabel("0"));
        graduations.put(new Integer(maxJour), new JLabel(""+maxJour));
        sliderJour.setLabelTable(graduations);
        sliderJour.setPaintLabels(true);
        sliderJour.setBackground(Color.white);
        sliderPane.add(sliderJour);
        
        mois = moisInitial;
        labelMois = new JLabel("Mois = " + mois);
        labelMois.setFont(gras);
        labelPane.add(labelMois);
        sliderMois = new JSlider(minMois, maxMois, mois);
        sliderMois.setSnapToTicks(true);
        sliderMois.addChangeListener(this);
        sliderMois.setMinorTickSpacing(1);
        sliderMois.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(minMois), new JLabel(""+minMois));
        graduations.put(new Integer(maxMois), new JLabel(""+maxMois));
        sliderMois.setLabelTable(graduations);
        sliderMois.setPaintLabels(true);
        sliderMois.setBackground(Color.white);
        sliderPane.add(sliderMois);
        
        annee = anneeInitiale;
        labelAnnee = new JLabel("Anne = " + annee);
        labelAnnee.setFont(gras);
        labelPane.add(labelAnnee);
        sliderAnnee = new JSlider(minAnnee, maxAnnee, annee);
        sliderAnnee.setSnapToTicks(true);
        sliderAnnee.addChangeListener(this);
        sliderAnnee.setMajorTickSpacing(10);
        sliderAnnee.setMinorTickSpacing(1);
        sliderAnnee.setPaintTicks(true);
        graduations = new Hashtable();
        graduations.put(new Integer(minAnnee), new JLabel(""+minAnnee));
        graduations.put(new Integer(2010), new JLabel("2010"));
        graduations.put(new Integer(maxAnnee), new JLabel(""+maxAnnee));
        sliderAnnee.setLabelTable(graduations);
        sliderAnnee.setPaintLabels(true);
        sliderAnnee.setBackground(Color.white);
        sliderPane.add(sliderAnnee);
        
        if (gras.canDisplay('\u03B3'))
            titreGamma = "\u03B3";
        else
            titreGamma = "Y";
        
        rotObservateur = new Matrice(3, 3);
        rotEquateur = new Matrice(3, 3);
        O = new Vecteur(0, 0, 0);
        Nord = new Vecteur(1, Math.PI / 2, -Math.PI / 2);
        Sud = new Vecteur(1, Math.PI / 2, Math.PI / 2);
        Est = new Vecteur(1, Math.PI / 2, Math.PI);
        Ouest = new Vecteur(1, Math.PI / 2, 0);
        Z = new Vecteur(1, 0, 0);
        
        majCroquis();
        
        validate();
        if (ic != null)
            ic.requestFocus();
    }
    
    public void stateChanged(ChangeEvent e) {
        Object adj = e.getSource();
        if (adj == sliderAlpha) {
            double nalpha = minAlpha + (maxAlpha - minAlpha) * (sliderAlpha.getValue() / 96.0);
            if (nalpha != alpha) {
                alpha = nalpha;
                labelAlpha.setText(titreAlpha + " = " + affichageAngleHMS(alpha));
                majCroquis();
            }
        } else if (adj == sliderDelta) {
            double ndelta = minDelta + (maxDelta - minDelta) * (sliderDelta.getValue() / 180.0);
            if (ndelta != delta) {
                delta = ndelta;
                labelDelta.setText(titreDelta + " = " + affichageAngleDMS(delta, false, false));
                majCroquis();
            }
        } else if (adj == sliderPhi) {
            double nphi = minPhi + (maxPhi - minPhi) * (sliderPhi.getValue() / 180.0);
            if (nphi != phi) {
                phi = nphi;
                labelPhi.setText(titrePhi + " = " + affichageAngleDMS(phi, false, false));
                majCroquis();
            }
        } else if (adj == sliderL) {
            double nL = minL + (maxL - minL) * (sliderL.getValue() / 180.0);
            if (nL != L) {
                L = nL;
                labelL.setText("L = " + affichageAngleDMS(L, true, false));
                majCroquis();
            }
        } else if (adj == sliderHeure) {
            int nHeure = sliderHeure.getValue();
            if (nHeure != heure) {
                heure = nHeure;
                labelHeure.setText("Heure = " + heure + "h TU");
                majCroquis();
            }
        } else if (adj == sliderJour) {
            int nJour = sliderJour.getValue();
            if (nJour != jour) {
                jour = nJour;
                labelJour.setText("Jour = " + jour);
                majCroquis();
            }
        } else if (adj == sliderMois) {
            int nMois = sliderMois.getValue();
            if (nMois != mois) {
                mois = nMois;
                labelMois.setText("Mois = " + mois);
                maxJour = joursParMois(mois, annee);
                sliderJour.setMaximum(maxJour);
                Hashtable graduations = new Hashtable();
                graduations.put(new Integer(0), new JLabel("0"));
                graduations.put(new Integer(maxJour), new JLabel(""+maxJour));
                sliderJour.setLabelTable(graduations);
                majCroquis();
            }
        } else if (adj == sliderAnnee) {
            int nAnnee = sliderAnnee.getValue();
            if (nAnnee != annee) {
                annee = nAnnee;
                labelAnnee.setText("Anne = " + annee);
                maxJour = joursParMois(mois, annee);
                sliderJour.setMaximum(maxJour);
                Hashtable graduations = new Hashtable();
                graduations.put(new Integer(0), new JLabel("0"));
                graduations.put(new Integer(maxJour), new JLabel(""+maxJour));
                sliderJour.setLabelTable(graduations);
                majCroquis();
            }
        }
    }
    
    protected void creerCroquis() {
        croquis = createImage(imgdimx, imgdimy);
        g2d = (Graphics2D)croquis.getGraphics();
        gras = new Font(getFont().getName(), Font.BOLD, 12);
        g2d.setFont(gras);
        g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        g2d.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);
    }
    
    protected void majCroquis() {
        // calcul du temps sidral
        // Jour Julien le 1/1/2000  0 h UTC : 2451544.5
        int jours = jour - 1; // jours couls depuis le 1/1/2000
        for (int i=2000; i<annee; i++)
            jours += joursParAn(i);
        for (int i=1; i<12; i++) {
            if (mois > i)
                jours += joursParMois(i, annee);
        }
        
        double T = (2451544.5 + jours - 2451545) / 36525.; // en siecles juliens depuis le 1/1/2000  12h
        // temps sidral moyen  Greenwich  0hTU :
        double GMST0 = 24110.54841 + 8640184.812866*T + 0.093104*Math.pow(T,2) - 6.2E-6*Math.pow(T,3); // en secondes
        GMST0 = GMST0 % 86400; // en s (24h = 86400s)
        double TSG = GMST0 + 1.00273790935 * (heure*3600); // en s
        TSG = TSG % 86400;
        double TSL = TSG * ((2*Math.PI)/86400) + L; // en radians
        H = TSL - alpha;
        if (H < 0)
            H += 2 * Math.PI;
        
        final double aobs = 0.2; // angle d'observation
        rotObservateur.set(0, 0, Math.cos(aobs));
        rotObservateur.set(0, 1, 0);
        rotObservateur.set(0, 2, -Math.sin(aobs));
        rotObservateur.set(1, 0, 0);
        rotObservateur.set(1, 1, 1);
        rotObservateur.set(1, 2, 0);
        rotObservateur.set(2, 0, Math.sin(aobs));
        rotObservateur.set(2, 1, 0);
        rotObservateur.set(2, 2, Math.cos(aobs));
        
        final double cphi = Math.PI / 2 - phi; // angle de l'quateur par rapport  l'horizon
        rotEquateur.set(0, 0, 1);
        rotEquateur.set(0, 1, 0);
        rotEquateur.set(0, 2, 0);
        rotEquateur.set(1, 0, 0);
        rotEquateur.set(1, 1, Math.cos(cphi));
        rotEquateur.set(1, 2, Math.sin(cphi));
        rotEquateur.set(2, 0, 0);
        rotEquateur.set(2, 1, -Math.sin(cphi));
        rotEquateur.set(2, 2, Math.cos(cphi));
        
        rotObsEq = rotObservateur.produit(rotEquateur);
        
        g2d.setColor(Color.white);
        g2d.fillRect(0, 0, imgdimx, imgdimy);
        cx = imgdimx / 2;
        cy = imgdimy / 2;
        rayon = Math.min(cx - 20, cy - 20); // rayon de la sphre en pixels
        
        g2d.setColor(new Color(230, 230, 255));
        g2d.fillOval(cx - rayon, cy - rayon, rayon * 2, rayon * 2);
        
        g2d.setColor(Color.darkGray);
        g2d.drawOval(cx - rayon, cy - rayon, rayon * 2, rayon * 2);
        
        final Vecteur M = new Vecteur(1, Math.PI / 2 - delta, Math.PI / 2 - H); // position de l'toile dans le repre de l'quateur
        if (M.get(2) < -Math.PI)
            M.set(2, M.get(2) + 2 * Math.PI); // pour avoir la 3me coordonne entre -pi et pi
        final Vecteur Mcart = M.spheriquesVersCartesiennes();
        final Vecteur McartHoz = rotEquateur.produit(Mcart); // rotation quateur -> horizon
        final Vecteur MsphHoz = McartHoz.cartesiennesVersSpheriques(); // coordonnes sphriques de M dans le repre de l'horizon
        final Point2D ptM = conversion(M, true);
        final Point2D ptO = new Point2D.Double(cx, cy);
        final Vecteur B = new Vecteur(1, Math.PI / 2, Math.PI / 2 - H); // repre quatorial
        if (B.get(2) < -Math.PI)
            B.set(2, B.get(2) + 2 * Math.PI);
        final Point2D ptB = conversion(B, true);
        final Point2D ptNord = conversion(Nord, false);
        final Vecteur P = new Vecteur(1, Math.PI / 2 - phi, -Math.PI / 2);
        final Point2D ptP = conversion(P, false);
        final Vecteur A = new Vecteur(1, Math.PI / 2, MsphHoz.get(2)); // repre de l'horizon
        final Vecteur pointVernal = new Vecteur(1, Math.PI / 2, Math.PI/2 - (H + alpha)); // repre quateur
        if (pointVernal.get(2) < -Math.PI)
            pointVernal.set(2, pointVernal.get(2) + 2 * Math.PI);
        Point2D pt;
        
        // calcul cercle de l'quateur cleste (derrire)
        final GeneralPath cheminEq1 = new GeneralPath();
        pt = conversion(new Vecteur(1, Math.PI / 2, 0), true);
        cheminEq1.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<=precisionCourbes; i++) {
            pt = conversion(new Vecteur(1, Math.PI / 2, - (i / (double)precisionCourbes) * Math.PI), true);
            cheminEq1.lineTo((float)pt.getX(), (float)pt.getY());
        }
        cheminEq1.closePath();
        
        // arc ZMA
        dessinArcZMA(Z, MsphHoz, A, false);
        
        // arc PMB
        dessinArcPMB(P, M, B, false);
        
        dessinEtoile(M, ptM, MsphHoz, false);
        
        dessinAngleDelta(B, M, false);
        
        dessinAngleHauteur(A, MsphHoz, false);
        
        // affichage cercle de l'quateur cleste (derrire)
        g2d.setColor(couleurEquateur);
        g2d.fill(cheminEq1);
        
        // angle H (angle horaire) (partie derrire l'horizon)
        dessinAngleHoraire(B, false);
        
        dessinAngleAlpha(pointVernal, B, false);
        
        // angle a (azimut) (partie derrire l'horizon)
        if (MsphHoz.get(1) > 1E-5 && MsphHoz.get(1) < Math.PI - 1E-5) // toile au znith: azimut indtermin
            dessinAngleAzimut(A, false);
        
        dessinHorizon();
        
        // calcul cercle de l'quateur cleste (devant)
        final GeneralPath cheminEq2 = new GeneralPath();
        pt = conversion(new Vecteur(1, Math.PI / 2, 0), true);
        cheminEq2.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<=precisionCourbes/2; i++) {
            pt = conversion(new Vecteur(1, Math.PI / 2, (i / (double)(precisionCourbes/2)) * Math.PI), true);
            cheminEq2.lineTo((float)pt.getX(), (float)pt.getY());
        }
        cheminEq2.closePath();
        // affichage cercle de l'quateur cleste (devant)
        g2d.setColor(couleurEquateur);
        g2d.fill(cheminEq2);
        g2d.setColor(couleurTexteEquateur);
        if (phi > 0) {
            pt = conversion(Ouest, false);
            g2d.rotate(phi - Math.PI/2, pt.getX(), pt.getY());
            g2d.drawString("Equateur cleste", (float)(pt.getX() - rayon * 0.75), (float)pt.getY() + 10);
            g2d.rotate(Math.PI/2 - phi, pt.getX(), pt.getY());
        } else {
            pt = conversion(Est, false);
            g2d.rotate(phi + Math.PI/2, pt.getX(), pt.getY());
            g2d.drawString("Equateur cleste", (float)(pt.getX() - rayon * 0.75), (float)pt.getY() - 5);
            g2d.rotate(-phi - Math.PI/2, pt.getX(), pt.getY());
        }
        
        dessinArcZMA(Z, MsphHoz, A, true);
        
        dessinArcPMB(P, M, B, true);
        
        dessinEtoile(M, ptM, MsphHoz, true);
        
        // angle H (angle horaire) (partie devant l'horizon)
        dessinAngleHoraire(B, true);
        
        dessinAngleAlpha(pointVernal, B, true);
        
        dessinAngleDelta(B, M, true);
        
        dessinAnglePhi(P);
        
        // angle a (azimut) (partie devant l'horizon)
        if (MsphHoz.get(1) > 1E-5 && MsphHoz.get(1) < Math.PI - 1E-5) // toile au znith: azimut indtermin
            dessinAngleAzimut(A, true);
        
        // angle h (hauteur du soleil)
        dessinAngleHauteur(A, MsphHoz, true);
        
        // points
        g2d.setColor(Color.black);
        
        // affichage des points fixes
        afficherPoint(O, false, true, "O");
        afficherPoint(Nord, false, true, "Nord");
        afficherPoint(Sud, false, true, "Sud");
        afficherPoint(Est, false, true, "Est");
        afficherPoint(Ouest, false, true, "Ouest");
        afficherPoint(Z, false, true, "Z");
        
        // affichage des points en mouvement
        afficherPoint(P, false, true, "P");
        
        //afficherPoint(B, true, true, "B");
        //afficherPoint(A, false, true, "A");
        //afficherPoint(M, true, true, "M");
        afficherPoint(pointVernal, true, true, titreGamma);
        
        ic.updateImage(croquis);
        
        
        String infos = "H=" + affichageAngleDMS(H, false, true) + "   h=" + affichageAngleDMS(Math.PI/2 - MsphHoz.get(1), false, false);
        if (MsphHoz.get(1) > 1E-5 && MsphHoz.get(1) < Math.PI - 1E-5) {
            double azimut = Math.PI/2 - A.get(2);
            if (azimut < 0)
                azimut += 2*Math.PI;
            infos += "   a=" + affichageAngleDMS(azimut, false, true);
        }
        
        labelInfos.setText(infos);
    }
    
    protected void dessinHorizon() {
        final GeneralPath cheminHorizon = new GeneralPath();
        Point2D pt = conversion(new Vecteur(1, Math.PI / 2, 0), false);
        cheminHorizon.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<precisionCourbes/2; i++) {
            pt = conversion(new Vecteur(1, Math.PI / 2, (i / (double)(precisionCourbes/2)) * (2 * Math.PI)), false);
            cheminHorizon.lineTo((float)pt.getX(), (float)pt.getY());
        }
        cheminHorizon.closePath();
        g2d.setColor(couleurHorizon);
        g2d.fill(cheminHorizon);
        g2d.setColor(couleurTexteHorizon);
        pt = conversion(Ouest, false);
        g2d.rotate(-0.1, pt.getX(), pt.getY());
        g2d.drawString("Horizon cleste", (float)(pt.getX() + rayon / 4), (float)pt.getY() + 20);
        g2d.rotate(0.1, pt.getX(), pt.getY());
    }
    
    protected void dessinEtoile(final Vecteur M, final Point2D ptM, final Vecteur MsphHoz, final boolean devant) {
        final boolean Mdevant = (Math.cos(M.get(2)) > 0);
        if (Mdevant == devant) {
            final double ex = ptM.getX();
            final double ey = ptM.getY();
            final double er = 7;
            double a, x, y;
            int[] ta = {2, 4, 1, 3};
            a = Math.PI / 2;
            x = ex + er * Math.cos(a);
            y = ey - er * Math.sin(a);
            final GeneralPath cheminEtoile = new GeneralPath();
            cheminEtoile.moveTo((float)x, (float)y);
            for (int i=0; i<ta.length; i++) {
                a = ta[i] * (2 * Math.PI / 5) + Math.PI / 2;
                x = ex + er * Math.cos(a);
                y = ey - er * Math.sin(a);
                cheminEtoile.lineTo((float)x, (float)y);
            }
            cheminEtoile.closePath();
            g2d.setColor(Color.black);
            g2d.draw(cheminEtoile);
            g2d.setColor(couleurEtoile);
            g2d.fill(cheminEtoile);
        }
    }
    
    protected void dessinArcZMA(final Vecteur Z, final Vecteur MsphHoz, final Vecteur A, final boolean devant) {
        final boolean Adevant = (Math.cos(A.get(2)) > 0);
        if (Adevant == devant) {
            g2d.setColor(couleurArcs);
            final Vecteur Zbis = new Vecteur(1, 0, MsphHoz.get(2)); // Z avec le mme phi que SsphHoz
            dessinArc(Zbis, MsphHoz.get(1) < A.get(1) ? MsphHoz : A, true, false, false);
        }
    }
    
    protected void dessinArcPMB(final Vecteur P, final Vecteur M, final Vecteur B, final boolean devant) {
        final boolean Bdevant = (Math.cos(B.get(2)) > 0);
        if (Bdevant == devant) {
            g2d.setColor(couleurArcs);
            final Vecteur PEq = new Vecteur(1, 0, M.get(2)); // P dans le repre quatorial, avec le mme phi que S
            dessinArc(PEq, M.get(1) < B.get(1) ? M : B, true, true, false);
        }
    }
    
    protected void dessinAngleAlpha(final Vecteur pointVernal, final Vecteur B, final boolean devant) {
        g2d.setColor(couleurAlpha);
        final boolean gammaDevant = (Math.cos(pointVernal.get(2)) > 0);
        final boolean gammaDerriere = (Math.cos(pointVernal.get(2)) <= 0);
        final boolean Bdevant = (Math.cos(B.get(2)) > 0);
        final boolean Bderriere = (Math.cos(B.get(2)) <= 0);
        // dbut angle
        if (devant && ((gammaDevant && Bderriere) || (alpha > Math.PI && gammaDevant && Bdevant))) {
            // trait gamma-Sud (devant)
            dessinArc(pointVernal, Sud, true, true, false);
        } else if (!devant && ((gammaDerriere && Bdevant) || (alpha > Math.PI && gammaDerriere && Bderriere))) {
            // trait gamma-Nord (derrire)
            dessinArc(pointVernal, Nord, true, true, false);
        }
        // milieu
        if (alpha > Math.PI) {
            if (devant && gammaDerriere && Bderriere) {
                // trait Nord-Ouest-Sud (devant)
                dessinArc(Nord, Sud, true, true, false);
            } else if (!devant && gammaDevant && Bdevant) {
                // trait Sud-Est-Nord (derrire)
                dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), new Vecteur(1, Math.PI/2, 3*Math.PI/2), true, true, false);
            }
        }
        // fin avec la flche
        if (devant && alpha < Math.PI && gammaDevant && Bdevant) {
            // flche gamma-B (devant)
            dessinArc(pointVernal, B, true, true, true);
        } else if (!devant && alpha < Math.PI && gammaDerriere && Bderriere) {
            // flche gamma-B (derrire)
            dessinArc(pointVernal, B, true, true, true);
        } else if (devant && ((gammaDerriere && Bdevant) || (alpha > Math.PI && gammaDevant && Bdevant))) {
            // flche Nord-B (devant)
            dessinArc(new Vecteur(1, Math.PI/2, -Math.PI/2), B, true, true, true);
        } else if (!devant && ((gammaDevant && Bderriere) || (alpha > Math.PI && gammaDerriere && Bderriere))) {
            // flche Sud-B (derrire)
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), B, true, true, true);
        }
        // titre
        if (devant) {
            double moyennePhi;
            if (B.get(2) - pointVernal.get(2) < 0)
                moyennePhi = pointVernal.get(2) + (B.get(2) + 2*Math.PI - pointVernal.get(2))/2;
            else
                moyennePhi = pointVernal.get(2) + (B.get(2) - pointVernal.get(2))/2;
            Point2D pt = conversion(new Vecteur(1, Math.PI/2, moyennePhi), true);
            if (phi > 0)
                g2d.drawString(titreAlpha, (float)(pt.getX() - 10), (float)(pt.getY() - 5));
            else
                g2d.drawString(titreAlpha, (float)(pt.getX()), (float)(pt.getY() - 5));
        }
    }
    
    protected void dessinAngleDelta(final Vecteur B, final Vecteur M, final boolean devant) {
        g2d.setColor(couleurDelta);
        final boolean Bdevant = (Math.cos(B.get(2)) > 0);
        if (devant == Bdevant) {
            dessinArc(B, M, true, true, true);
            if (Math.abs(delta) > 0.1) {
                double r;
                if ((phi >= 0 && delta >= 0) || (phi < 0 && delta < 0))
                    r = 1.1;
                else
                    r = 0.9;
                Point2D pt = conversion(new Vecteur(r, Math.PI/2 - (Math.PI/2 - M.get(1))/2, B.get(2)), true);
                g2d.drawString(titreDelta, (float)(pt.getX() - g2d.getFontMetrics().stringWidth(titreDelta)/2), (float)(pt.getY() + 5));
            }
        }
    }
    
    protected void dessinAnglePhi(final Vecteur P) {
        g2d.setColor(couleurPhi);
        dessinArc(Nord, P, true, false, true);
        Point2D pt = conversion(new Vecteur(1.05, Math.PI/2 - phi/2, -Math.PI/2), false);
        g2d.drawString(titrePhi, (float)(pt.getX() - g2d.getFontMetrics().stringWidth(titrePhi)/2), (float)(pt.getY() + 5));
    }
    
    protected void dessinAngleAzimut(final Vecteur A, final boolean devant) {
        double azimut = Math.PI/2 - A.get(2);
        if (azimut < 0)
            azimut += 2*Math.PI;
        final Vecteur Az = new Vecteur(1, Math.PI / 2, Math.PI/2 - azimut);
        if ((devant && azimut <= Math.PI) || (!devant && azimut > Math.PI)) {
            g2d.setColor(couleurAngleA);
            dessinArc(Sud, Az, false, false, true);
        } else if (devant && azimut > Math.PI) {
            g2d.setColor(couleurAngleA);
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), new Vecteur(1, Math.PI/2, -Math.PI/2), false, false, false);
        }
        if (devant) {
            Point2D pt = conversion(new Vecteur(1, Math.PI/2, Math.PI/2 - 0.75*azimut), false);
            g2d.drawString("a", (float)(pt.getX() - 10), (float)(pt.getY() - 5));
        }
    }
    
    protected void dessinAngleHoraire(final Vecteur B, final boolean devant) {
        if ((devant && H <= Math.PI) || (!devant && H > Math.PI)) {
            g2d.setColor(couleurAngleH);
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), B, false, true, true);
        } else if (devant && H > Math.PI) {
            g2d.setColor(couleurAngleH);
            dessinArc(new Vecteur(1, Math.PI/2, Math.PI/2), new Vecteur(1, Math.PI/2, -Math.PI/2), false, true, false);
        }
        if (devant) {
            Point2D pt = conversion(new Vecteur(1, Math.PI/2, Math.PI/2 - 0.75 * H), true);
            g2d.drawString("H", (float)(pt.getX() - 10), (float)(pt.getY() - 5));
        }
    }
    
    protected void dessinAngleHauteur(final Vecteur A, final Vecteur MsphHoz, final boolean devant) {
        final boolean Adevant = (Math.cos(A.get(2)) > 0);
        if (Adevant == devant) {
            g2d.setColor(couleurAngleHauteur);
            dessinArc(A, MsphHoz, true, false, true);
            if (Math.abs(MsphHoz.get(1) - Math.PI/2) > 0.1) {
                final Point2D pt = conversion(new Vecteur(1, Math.PI/2 - (Math.PI/2 - MsphHoz.get(1))/2, A.get(2)), false);
                double x;
                if (delta >= 0)
                    x = pt.getX() - 10;
                else
                    x = pt.getX() + 5;
                g2d.drawString("h", (float)(x), (float)(pt.getY() + 5));
            }
        }
    }
    
    // dessin d'un arc entre 2 points dans l'un des 2 systmes de coordonnes sphriques, en suivant l'une des coordonnes
    protected void dessinArc(final Vecteur v1, final Vecteur v2, final boolean sensPhiPositif, final boolean coordEq, final boolean fleche) {
        final int nb = precisionCourbes/4;
        final double dr = (v2.get(0) - v1.get(0)) / nb;
        final double dtheta = (v2.get(1) - v1.get(1)) / nb;
        double phi1 = v1.get(2);
        double phi2 = v2.get(2);
        if (v1.get(1) == 0)
            phi1 = phi2;
        else if (v2.get(1) == 0)
            phi2 = phi1;
        if (sensPhiPositif && phi1 > phi2)
            phi2 += 2 * Math.PI;
        else if (!sensPhiPositif && phi1 < phi2)
            phi1 += 2 * Math.PI;
        final double dphi = (phi2 - phi1) / nb; // ne marche pas toujours (par ex. avec theta = 0)
        final GeneralPath chemin = new GeneralPath();
        Point2D pt = conversion(v1, coordEq);
        Point2D vpt = null;
        chemin.moveTo((float)pt.getX(), (float)pt.getY());
        for (int i=1; i<=nb; i++) {
            if (i == nb)
                vpt = pt;
            pt = conversion(new Vecteur(v1.get(0) + i * dr, v1.get(1) + i * dtheta, phi1 + i * dphi), coordEq);
            chemin.lineTo((float)pt.getX(), (float)pt.getY());
        }
        g2d.draw(chemin);
        if (fleche && Math.abs(v2.get(0) - v1.get(0)) + Math.abs(v2.get(1) - v1.get(1)) + Math.abs(phi2 - phi1) > 0)
            dessinFleche(vpt, pt);
    }
    
    // dessin des 2 traits de la fin de la flche  partir des 2 derniers points d'un arc
    protected void dessinFleche(final Point2D pt1, final Point2D pt2) {
        final double l = 6; // longueur des traits en pixels
        final double da = Math.PI / 8; // angle des traits avec la fin de l'arc
        final double dx = pt2.getX() - pt1.getX();
        final double dy = pt2.getY() - pt1.getY();
        double angle;
        if (dx == 0) {
            if (dy > 0)
                angle = Math.PI / 2;
            else
                angle = -Math.PI / 2;
        } else {
            angle = Math.atan(dy / dx);
            if (dx < 0)
                angle = angle + Math.PI;
        }
        Point2D ptFleche = new Point2D.Double(pt2.getX() - l * Math.cos(angle + da), pt2.getY() - l * Math.sin(angle + da));
        g2d.draw(new Line2D.Double(pt2.getX(), pt2.getY(), ptFleche.getX(), ptFleche.getY()));
        ptFleche = new Point2D.Double(pt2.getX() - l * Math.cos(angle - da), pt2.getY() - l * Math.sin(angle - da));
        g2d.draw(new Line2D.Double(pt2.getX(), pt2.getY(), ptFleche.getX(), ptFleche.getY()));
    }
    
    // passage des coordonnes sphriques aux coordonnes cran
    protected Point2D conversion(final Vecteur vsph, final boolean rotEq) {
        // passage aux coordonnes cartsiennes
        Vecteur v = vsph.spheriquesVersCartesiennes();
        
        // rotation optionelle de l'quateur et rotation permettant de voir l'horizon de haut
        if (rotEq)
            v = rotObsEq.produit(v);
        else
            v = rotObservateur.produit(v);
        
        return(projectionEcran(v));
    }
    
    // ajout d'un point en coordonnes sphriques (r, theta, phi)
    protected void afficherPoint(final Vecteur vsph, final boolean rotEq, final boolean croix, final String titre) {
        final Point2D p = conversion(vsph, rotEq);
        
        // affichage d'une croix ou d'un point
        if (croix) {
            g2d.draw(new Line2D.Double(p.getX() - 2, p.getY(), p.getX() + 2, p.getY()));
            g2d.draw(new Line2D.Double(p.getX(), p.getY() - 2, p.getX(), p.getY() + 2));
        } else
            g2d.draw(new Line2D.Double(p.getX(), p.getY(), p.getX(), p.getY()));
        
        // affichage du titre
        if (titre != null) {
            final int l = g2d.getFontMetrics().stringWidth(titre);
            g2d.drawString(titre, (float)p.getX() - l/2, (float)p.getY() - 9);
        }
    }
    
    // projection selon l'axe x  partir des coordonnes cartsiennes, puis conversion vers l'cran
    protected Point2D projectionEcran(final Vecteur v) {
        return(new Point2D.Double(cx + v.get(1) * rayon, cy - v.get(2) * rayon));
    }
    
    // affichage en heures minutes d'un angle en radians > 0
    protected static String affichageAngleHMS(final double a) {
        double h = a*12/Math.PI;
        int ih = (int)Math.floor(h);
        double m = (h - ih)*60;
        int im = (int)Math.round(m);
        if (im >= 60) {
            im -= 60;
            ih += 1;
        }
        return(ih + "h " + im + "m");
    }
    
    // affichage en degrs minutes d'un angle en radians
    protected static String affichageAngleDMS(final double a, final boolean estOuest, final boolean angle360) {
        double d = Math.toDegrees(Math.abs(a));
        int id = (int)Math.floor(d);
        double m = (d - id)*60;
        int im = (int)Math.round(m);
        if (im >= 60) {
            im -= 60;
            id += 1;
        }
        String signe;
        if (angle360 || a == 0)
            signe = "";
        else if (a > 0)
            signe = "+";
        else
            signe = "-";
        if (estOuest)
            return(id + " " + im + "'" + (a >= 0 ? " E" : " O"));
        else
            return(signe + id + " " + im + "'");
    }
    
    protected static int joursParMois(final int mois, final int an) {
        switch (mois) {
            case 1: return(31);
            case 2: {
                if (an % 4 == 0 && (an % 100 != 0 || an % 400 == 0))
                    return(29);
                else
                    return(28);
            }
            case 3: return(31);
            case 4: return(30);
            case 5: return(31);
            case 6: return(30);
            case 7: return(31);
            case 8: return(31);
            case 9: return(30);
            case 10: return(31);
            case 11: return(30);
            case 12: return(31);
        }
        return(0);
    }
    
    protected static int joursParAn(final int an) {
        if (an % 4 == 0 && (an % 100 != 0 || an % 400 == 0))
            return(366);
        else
            return(365);
    }
}
